/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponce
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor.specific;

import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.dnd.DragSource;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JOptionPane;

import ch.cern.lhcb.xmleditor.Editor;
import ch.cern.lhcb.xmleditor.ResourceManager;

/**
 * This manages a set of icons associated to names
 * 
 * @author Sebastien Ponce
 */
public class IconsManager {

    /**
     * a hashtable containing all icons used here
     */
    private Hashtable icons;

    /**
     * the underlying Editor
     */
    private Editor editor;

    /**
     * default constructor it basically populates the IconsManager with default
     * icons
     * 
     * @param editor the underlying Editor
     */
    public IconsManager(Editor editor) {
        super();
        this.editor = editor;
        icons = new Hashtable();
        // gets the icons names and retrieves the images
        fillIconTable();
    }

    /**
     * retrieves all the icon images and assocites them to elements if some icon
     * files are not valid, they are simply ignored
     */
    private void fillIconTable() {
        // retrieves the icon file names from the Setup
        Map map = editor.getSetup().getIcons();
        if (map == null) { return; }
        // this will contain the name of non-existing icon file names
        Vector nonExistingFiles = new Vector();
        // for each icon, try to add it to the icon table
        Iterator it = map.keySet().iterator();
        while (it.hasNext()) {
            String elemName = (String) it.next();
            String fileName = (String) map.get(elemName);
            URL fileURL = ClassLoader.getSystemResource(ResourceManager
                    .getString("icons_url")
                    + fileName);
            if (fileURL != null) {
                icons.put(elemName, new ImageIcon(fileURL));
            } else {
                // the icon filename does probably not exists, adds it to the
                // list
                // of non-existing files
                nonExistingFiles.add(fileName);
            }
        }
        // if nonExistingFiles is not empty, displays an error message
        if (!nonExistingFiles.isEmpty()) {
            String message = "Unable to open the following icon files :\n";
            Enumeration files = nonExistingFiles.elements();
            while (files.hasMoreElements()) {
                message = message + "    " + (String) files.nextElement()
                        + "\n";
            }
            JOptionPane.showMessageDialog(editor, message);
        }
    }

    /**
     * forces the icon set to be reloaded from the current setup
     */
    public void reload() {
        icons.clear();
        fillIconTable();
    }

    /**
     * returns an icon out of a name
     * 
     * @param name a name designating the icon
     */
    public Icon getIcon(String name) {
        return (Icon) icons.get(name);
    }

    /**
     * returns a cursor out of a name
     * 
     * @param name a name designating the cursor
     * @return the cursor
     */
    public Cursor getCursor(String name) {
        ImageIcon image = (ImageIcon) icons.get(name);
        if (image != null) {
            try {
                Toolkit toolkit = java.awt.Toolkit.getDefaultToolkit();
                Dimension cursorDimension = toolkit.getBestCursorSize(image
                        .getIconWidth(), image.getIconHeight());
                return toolkit.createCustomCursor(image.getImage(), new Point(
                        cursorDimension.width / 2, cursorDimension.height / 2),
                        name);
            } catch (IndexOutOfBoundsException e) {
                // This means the cursor was smaller than expected
                // In this case, put the Point in the top left corner
                return java.awt.Toolkit.getDefaultToolkit().createCustomCursor(
                        image.getImage(), new Point(0, 0), name);
            }
        } else {
            return null;
        }
    }

    /**
     * returns a cursor for drag and drop in case the action is copy
     * 
     * @param isDropOk says whether the drop is ok or not
     * @return the cursor
     */
    public Cursor getCopyDropCursor(boolean isDropOk) {
        Cursor result;
        if (isDropOk) {
            result = this.getCursor("copyDrop");
            if (result == null) {
                result = DragSource.DefaultCopyDrop;
            }
        } else {
            result = this.getCursor("copyNoDrop");
            if (result == null) {
                result = DragSource.DefaultCopyNoDrop;
            }
        }
        return result;
    }

    /**
     * returns a cursor for drag and drop in case the action is move
     * 
     * @param isDropOk says whether the drop is ok or not
     * @return the cursor
     */
    public Cursor getMoveDropCursor(boolean isDropOk) {
        Cursor result;
        if (isDropOk) {
            result = this.getCursor("moveDrop");
            if (result == null) {
                result = DragSource.DefaultMoveDrop;
            }
        } else {
            result = this.getCursor("moveNoDrop");
            if (result == null) {
                result = DragSource.DefaultMoveNoDrop;
            }
        }
        return result;
    }

    /**
     * returns a cursor for drag and drop in case the action is link
     * 
     * @param isDropOk says whether the drop is ok or not
     * @return the cursor
     */
    public Cursor getLinkDropCursor(boolean isDropOk) {
        Cursor result;
        if (isDropOk) {
            result = this.getCursor("linkDrop");
            if (result == null) {
                result = DragSource.DefaultLinkDrop;
            }
        } else {
            result = this.getCursor("linkNoDrop");
            if (result == null) {
                result = DragSource.DefaultLinkNoDrop;
            }
        }
        return result;
    }

    /**
     * returns a cursor for drag and drop in case no action is allowed
     * 
     * @return the cursor
     */
    public Cursor getNoDropCursor() {
        Cursor result = this.getCursor("noDrop");
        if (result == null) {
            result = DragSource.DefaultMoveNoDrop;
        }
        return result;
    }

}