/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponce
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor.specific;

import java.text.Collator;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.w3c.dom.Node;

import ch.cern.lhcb.xmleditor.xerces.ElementImpl;
import ch.cern.lhcb.xmleditor.xerces.TreeElement;

import com.wutka.dtd.DTD;
import com.wutka.dtd.DTDAny;
import com.wutka.dtd.DTDAttribute;
import com.wutka.dtd.DTDChoice;
import com.wutka.dtd.DTDContainer;
import com.wutka.dtd.DTDDecl;
import com.wutka.dtd.DTDElement;
import com.wutka.dtd.DTDEmpty;
import com.wutka.dtd.DTDEnumeration;
import com.wutka.dtd.DTDItem;
import com.wutka.dtd.DTDMixed;
import com.wutka.dtd.DTDName;
import com.wutka.dtd.DTDNotationList;
import com.wutka.dtd.DTDPCData;
import com.wutka.dtd.DTDSequence;

/**
 * This describes the way Xml files are displayed
 * 
 * @author Sebastien Ponce
 */
public class XmlDisplay {

    /**
     * the internal Dtd
     */
    private DTD internalDtd;

    /**
     * the file containing the external dtd if any
     */
    private DtdFile dtdFile;

    /**
     * the Hashtable of the elements declared in dtd, associated to their name
     */
    private Hashtable elements;

    /**
     * the ElementName defining the display of text nodes
     */
    private ElementName textNodeName;

    /**
     * the ElementName defining the display of comments
     */
    private ElementName commentName;

    /**
     * constructor
     * 
     * @param dtdFile the external dtd associated with this display manager
     * @param internalDtd the internal dtd associated with this display manager
     */
    public XmlDisplay(DtdFile dtdFile, DTD internalDtd) {
        // initializes the fields
        this.internalDtd = internalDtd;
        this.dtdFile = dtdFile;
        elements = new Hashtable();
        textNodeName = new ElementName(ElementName.VALUE_NAME, "");
        commentName = new ElementName(ElementName.VALUE_NAME, "");
        // here, the dtd is analysed to build a list of elements
        // present in it and for each, the list of its attributes
        // and subelements
        if (internalDtd != null) {
            Enumeration internalDtdElements = internalDtd.elements.elements();
            fillElements(internalDtdElements);
        }
        if (dtdFile != null) {
            Enumeration extDtdElements = dtdFile.getDtd().elements.elements();
            fillElements(extDtdElements);
        }
        // avoid display of attributes name and href
        // since they are used as name of the node
        this.setAttributeDisplay("name", false);
        this.setAttributeDisplay("href", false);
        // and no more as element in the tree
        this.displaySubElementAsAttr();
        // initializes the names of the different elements
        Enumeration enm = elements.elements();
        while (enm.hasMoreElements()) {
            ((Element) enm.nextElement()).initializeName();
        }
    }

    /**
     * Fills the table of declared elements using the elements contained in
     * dtdElements
     * 
     * @param dtdElements an enumeration od DTDElements to add to the currently
     *        defined elements
     */
    private void fillElements(Enumeration dtdElements) {
        while (dtdElements.hasMoreElements()) {
            // gets an element out of the DTD
            DTDElement dtdElement = (DTDElement) dtdElements.nextElement();
            // creates a next Element in the DisplayManager
            Element element = new Element();
            // deals with attributes
            Enumeration attributes = dtdElement.attributes.elements();
            while (attributes.hasMoreElements()) {
                // gets attributes one by one
                DTDAttribute dtdAttribute = (DTDAttribute) attributes
                        .nextElement();
                // the default value
                String defaultValue = dtdAttribute.getDefaultValue();
                // depending on their nature, creates an Attribute with or
                // without
                // possible values
                Attribute attribute;
                if (dtdAttribute.type instanceof String) {
                    attribute = new Attribute(null, dtdAttribute.getDecl(),
                            defaultValue);
                } else if (dtdAttribute.type instanceof DTDEnumeration) {
                    attribute = new Attribute(
                            ((DTDEnumeration) dtdAttribute.type).getItems(),
                            dtdAttribute.getDecl(), defaultValue);
                } else if (dtdAttribute.type instanceof DTDNotationList) {
                    attribute = new Attribute(
                            ((DTDNotationList) dtdAttribute.type).getItems(),
                            dtdAttribute.getDecl(), defaultValue);
                } else {
                    // should never be reached
                    attribute = null;
                }
                // adds the new Attribute to the element
                element.addAttribute(dtdAttribute.name, attribute);
            }
            // deals with subElements
            DTDItem item = dtdElement.getContent();
            element.addItem(item);
            // adds the new element to the hashtable
            elements.put(dtdElement.getName(), element);
        }
    }

    /**
     * says whether the element with name elementName has a text value or not.
     * 
     * @param elementName element name
     * @return true if it has a text value
     */
    public boolean elementHasValue(String elementName) {
        Element element = (Element) elements.get(elementName);
        if (element != null) { return element.hasValue(); }
        // in all bad cases, ie if the dtd does not give the answer, returns
        // false
        return false;
    }

    /**
     * says whether the Attribute attributeName of Element elementName should be
     * displayed
     * 
     * @param elementName the name of the element concerned
     * @param attributeName the name of the attribute concerned
     * @return whether this attribute is displayed or not
     */
    public boolean isAttributeDisplayed(String elementName, String attributeName) {
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            Attribute attribute = (Attribute) element.attributes
                    .get(attributeName);
            if (attribute != null) { return attribute.getDisplayed(); }
        }
        // in all bad cases, ie if the dtd does not give the answer, returns
        // true
        return true;
    }

    /**
     * says whether the Attribute attributeName of Element elementName is
     * required or not
     * 
     * @param elementName the name of the element concerned
     * @param attributeName the name of the attribute concerned
     * @return whether this attribute is required
     */
    public boolean isAttributeRequired(String elementName, String attributeName) {
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            Attribute attribute = (Attribute) element.attributes
                    .get(attributeName);
            if (attribute != null) { return attribute.isRequired(); }
        }
        // in all bad cases, ie if the dtd does not give the answer, returns
        // false
        return false;
    }

    /**
     * says whether the Attribute attributeName of Element elementName is
     * editable or not
     * 
     * @param elementName the name of the element concerned
     * @param attributeName the name of the attribute concerned
     * @return whether this attribute is editable
     */
    public boolean isAttributeEditable(String elementName, String attributeName) {
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            Attribute attribute = (Attribute) element.attributes
                    .get(attributeName);
            if (attribute != null) { return attribute.isEditable(); }
        }
        // in all bad cases, ie if the dtd does not give the answer, returns
        // false
        return false;
    }

    /**
     * gets the possible values of Attribute attributeName of Element
     * elementName
     * 
     * @param elementName the name of the element concerned
     * @param attributeName the name of the attribute concerned
     * @return the possible values or null if any value is possible
     */
    public String[] getAttributePossibleValues(String elementName,
            String attributeName) {
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            Attribute attribute = (Attribute) element.attributes
                    .get(attributeName);
            if (attribute != null) { return attribute.getPossibleValues(); }
        }
        // in all bad cases, ie if the dtd does not give the answer, returns
        // null
        return null;
    }

    /**
     * gets the default value of Attribute attributeName of Element elementName
     * 
     * @param elementName the name of the element concerned
     * @param attributeName the name of the attribute concerned
     * @return the default value (possibly "")
     */
    public String getAttributeDefaultValue(String elementName,
            String attributeName) {
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            Attribute attribute = (Attribute) element.attributes
                    .get(attributeName);
            if (attribute != null) { return attribute.getDefaultValue(); }
        }
        // in all bad cases, ie if the dtd does not give the answer, returns a
        // default value
        return "aValue";
    }

    /**
     * if isDisplayed is true, the Attribute attributeName of Element
     * elementName will be displayed, if it is false, it will be hidden
     * 
     * @param elementName the name of the element concerned
     * @param attributeName the name of the attribute concerned
     * @param isDisplayed whether this attribute must be displayed or not
     */
    public void setAttributeDisplayed(String elementName, String attributeName,
            boolean isDisplayed) {
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            Attribute attribute = (Attribute) element.attributes
                    .get(attributeName);
            if (attribute != null) {
                attribute.setDisplayed(isDisplayed);
            }
        }
    }

    /**
     * manages the displaying of attributes with the given name
     * 
     * @param name the name of attributes that is managed
     * @param displayed true to display it, false to hide it
     */
    public void setAttributeDisplay(String name, boolean displayed) {
        // for each element
        Enumeration enm = this.elements.elements();
        while (enm.hasMoreElements()) {
            // looks whether tha attribute named name exists
            Element elem = (Element) enm.nextElement();
            Attribute attr = (Attribute) elem.attributes.get(name);
            // if yes
            if (attr != null) {
                // sets its display
                attr.setDisplayed(displayed);
            }
        }
    }

    /**
     * says whether the subelement subElemName of Element elementName should be
     * displayed
     * 
     * @param elementName the name of the element concerned
     * @param subElemName the name of the subelement concerned
     * @return whether this subelement is displayed or not
     */
    public boolean isSubElementDisplayed(String elementName, String subElemName) {
        // gets the element
        Element element = (Element) elements.get(elementName);
        // if unknown in the dtd, displays it
        if (element == null) { return true; }
        // gets the subelement
        SubElement subElement = (SubElement) element.subElements
                .get(subElemName);
        // if unknown in the dtd
        if (subElement == null) {
            // first look whether the corresponding reference is in the dtd
            SubElement refSubElement = (SubElement) element.subElements
                    .get(subElemName + "ref");
            // if yes, use the reference subelement
            if (refSubElement != null) {
                return refSubElement.getDisplayed();
            } else {
                // else unknown subelement, display it
                return true;
            }
        }
        // looks whether to display it
        return subElement.getDisplayed();
    }

    /**
     * says whether the subelement subElemName of Element elementName should be
     * displayed as attribute of its parent
     * 
     * @param elementName the name of the element concerned
     * @param subElemName the name of the subelement concerned
     * @return whether this subelement is displayed as attribtue of its parent
     *         or not
     */
    public boolean isSubElementDisplayedAsAttr(String elementName,
            String subElemName) {
        // gets the element
        Element element = (Element) elements.get(elementName);
        // if unknown in the dtd, don't display it
        if (element == null) { return false; }
        // gets the subelement
        SubElement subElement = (SubElement) element.subElements
                .get(subElemName);
        // if unknown in the dtd, don't display it
        if (subElement == null) { return false; }
        // looks whether to display it
        return subElement.getDisplayedAsAttr();
    }

    /**
     * if isDisplayed is true, the SubElement subElemName of Element elementName
     * will be displayed in the tree, if it is false, it will be hidden
     * 
     * @param elementName the name of the element concerned
     * @param subElemName the name of the subelement concerned
     * @param isDisplayed whether this subelement must be displayed or not
     */
    public void setSubElementDisplayed(String elementName, String subElemName,
            boolean isDisplayed) {
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            SubElement subElement = (SubElement) element.subElements
                    .get(subElemName);
            if (subElement != null) {
                subElement.setDisplayed(isDisplayed);
            }
        }
    }

    /**
     * if isDisplayed is true, the SubElement subElemName of Element elementName
     * will be displayed as an attribute, if it is false, it will not
     * 
     * @param elementName the name of the element concerned
     * @param subElemName the name of the subelement concerned
     * @param isDisplayed whether this subelement must be displayed or not
     */
    public void setSubElementDisplayedAsAttr(String elementName,
            String subElemName, boolean isDisplayed) {
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            SubElement subElement = (SubElement) element.subElements
                    .get(subElemName);
            if (subElement != null) {
                subElement.setDisplayedAsAttr(isDisplayed);
            }
        }
    }

    /**
     * display all subElement that are attribute like as attribute
     */
    public void displaySubElementAsAttr() {
        Enumeration enm = elements.elements();
        while (enm.hasMoreElements()) {
            Element element = (Element) enm.nextElement();
            Enumeration subelements = element.subElements.keys();
            while (subelements.hasMoreElements()) {
                String name = (String) subelements.nextElement();
                Element subelemAsElem = (Element) elements.get(name);
                if (subelemAsElem != null) {
                    if (subelemAsElem.isAttributeLike()) {
                        SubElement subelement = (SubElement) element.subElements
                                .get(name);
                        subelement.setDisplayed(false);
                        subelement.setDisplayedAsAttr(true);
                    }
                }
            }
        }
    }

    /**
     * says wether the given element could be seen as an attribute or not To be
     * seen as an attribute means that it only has one subelement (and no
     * attribute) and that this subelement is a text one, containing a value
     * 
     * @param name the name of the element
     * @return true if the element is attribute like
     */
    public boolean isElementAttributeLike(String name) {
        // tries to get the element (it could not exist if it is PCDATA for
        // example)
        Element element = (Element) elements.get(name);
        // if it doesn't exist, it is not attribute like
        if (element == null) {
            return false;
        } else {
            // else, asks the element to give the answer
            return element.isAttributeLike();
        }
    }

    /**
     * manages the displaying of subelements with the given name
     * 
     * @param name the name of subelements that shouldn't be displayed
     * @param displayed true to display it, false to hide it
     */
    public void setSubElementDisplay(String name, boolean displayed) {
        // for each element
        Enumeration enm = this.elements.elements();
        while (enm.hasMoreElements()) {
            // looks whether the subelement named name exists
            Element elem = (Element) enm.nextElement();
            SubElement subElem = (SubElement) elem.subElements.get(name);
            // if yes
            if (subElem != null) {
                // sets its display
                subElem.setDisplayed(displayed);
            }
        }
    }

    /**
     * manages the displaying of subelements with the given name as attributes
     * of their parents
     * 
     * @param name the name of subelements that shouldn't be displayed as
     *        attribute of its parent
     * @param displayed true to display it as attribute of its parent, false to
     *        hide it
     */
    public void setSubElementDisplayAsAttr(String name, boolean displayed) {
        // for each element
        Enumeration enm = this.elements.elements();
        while (enm.hasMoreElements()) {
            // looks whether the subelement named name exists
            Element elem = (Element) enm.nextElement();
            SubElement subElem = (SubElement) elem.subElements.get(name);
            // if yes
            if (subElem != null) {
                // sets its display
                subElem.setDisplayedAsAttr(displayed);
            }
        }
    }

    /**
     * gives all the element names of the underlying dtd, in alphabetical order
     * 
     * @return an array of names
     */
    public String[] getElementNames() {
        String[] result = new String[elements.size()];
        Enumeration enm = elements.keys();
        int i = 0;
        while (enm.hasMoreElements()) {
            result[i] = (String) enm.nextElement();
            i++;
        }
        Arrays.sort(result, Collator.getInstance());
        return result;
    }

    /**
     * gives all the attribute names of a given element, in alphabetical order
     * 
     * @param elementName the element name
     * @return an array of names
     */
    public String[] getAttributeNames(String elementName) {
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            // gets all attributes
            Hashtable attrs = element.attributes;
            // puts the result in a string array
            String[] result = new String[attrs.size()];
            Enumeration enm = attrs.keys();
            int i = 0;
            while (enm.hasMoreElements()) {
                result[i] = (String) enm.nextElement();
                i++;
            }
            Arrays.sort(result, Collator.getInstance());
            return result;
        }
        return new String[0];
    }

    /**
     * gives the attribute names of required attributes for a given element
     * 
     * @param elementName the element name
     * @return an array of names
     */
    public String[] getRequiredAttributeNames(String elementName) {
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            // gets all attributes
            Hashtable attrs = element.attributes;
            // keeps those which are required
            Vector names = new Vector();
            Enumeration enm = attrs.keys();
            while (enm.hasMoreElements()) {
                String item = (String) enm.nextElement();
                if (this.isAttributeRequired(elementName, item)) {
                    names.add(item);
                }
            }
            // puts the result in a string array
            String[] result = new String[names.size()];
            Enumeration enm2 = names.elements();
            int i = 0;
            while (enm2.hasMoreElements()) {
                result[i] = (String) enm2.nextElement();
                i++;
            }
            return result;
        }
        return new String[0];
    }

    /**
     * gives all the subelements names of a given element
     * 
     * @param elementName the element name
     * @return an array of names
     */
    public String[] getSubElementNames(String elementName) {
        // first gets the element
        Element element = (Element) elements.get(elementName);
        if (element != null) {
            // if it accepts any subelement, first check that the
            // subelement table is initialized
            if (element.acceptAnySubElement) {
                if (element.subElements.isEmpty()) {
                    element.initializeAnySubElements();
                }
            }
            Hashtable subElems = element.subElements;
            String[] result = new String[subElems.size()];
            Enumeration enm = subElems.keys();
            int i = 0;
            while (enm.hasMoreElements()) {
                result[i] = (String) enm.nextElement();
                i++;
            }
            return result;
        }
        return new String[0];
    }

    /**
     * gets the ElementName used for text nodes
     * 
     * @return the elementName used for text nodes
     */
    public ElementName getTextNodeName() {
        return textNodeName;
    }

    /**
     * sets the elementName used for text nodes
     * 
     * @param textNodeName the new elementName
     */
    public void setTextNodeName(ElementName textNodeName) {
        this.textNodeName = textNodeName;
    }

    /**
     * gets the ElementName used for comments
     * 
     * @return the elementName used for comments
     */
    public ElementName getCommentName() {
        return commentName;
    }

    /**
     * sets the elementName used for comments
     * 
     * @param commentName the new elementName
     */
    public void setCommentName(ElementName commentName) {
        this.commentName = commentName;
    }

    /**
     * says whether the element named name exists here or not
     * 
     * @param name the name to test
     * @return true if the name exists in the dtd
     */
    public boolean isInDtd(String name) {
        Element element = (Element) elements.get(name);
        return element != null;
    }

    /**
     * gets the elementName of a given element
     * 
     * @param name the name of the element whose elementName we want to get
     * @return the elementName of this element
     */
    public ElementName getElementName(String name) {
        Element element = (Element) elements.get(name);
        if (element != null) { return element.getElementName(); }
        // in all bad cases, ie if the dtd does not give the answer, returns a
        // default ElementName
        return new ElementName(ElementName.TYPE_NAME, null);
    }

    /**
     * sets the elementName of a given element
     * 
     * @param name the name of the element which elementName will change
     * @param elementName the new elementName
     */
    public void setElementName(String name, ElementName elementName) {
        Element element = (Element) elements.get(name);
        if (element != null) {
            element.setElementName(elementName);
        }
    }

    /**
     * gets the ElementName from an Node Actually, it does only call the
     * getElementName method of the corresponding Element if this one exists. If
     * not, it builds a default ElementName.
     * 
     * @param node the Node considered
     * @return the ElementName to display the element
     */
    private ElementName getElementName(Node node) {
        switch (node.getNodeType()) {
        case Node.COMMENT_NODE:
            return commentName;
        case Node.TEXT_NODE:
            return textNodeName;
        case Node.ELEMENT_NODE:
            Element element = (Element) elements.get(node.getNodeName());
            // take care of a possibly null element
            if (element != null) {
                // if the element was not null, gets its elementName
                return element.getElementName();
            }
        }
        // else uses the default policy
        return getDefaultElementName(node);
    }

    /**
     * sets the name of a given Node
     * 
     * @param node the node concerned
     * @param name the new name
     */
    public void setNodeName(TreeElement node, String name) {
        // gets the ElementName for this node
        ElementName elementName = this.getElementName(node);
        switch (elementName.nameType) {
        // depending on its value, sets the right value
        case ElementName.TYPE_NAME:
            // this should never arrive
            // does nothing
            break;
        case ElementName.VALUE_NAME:
            node.setValue(name);
            break;
        case ElementName.ATTRIBUTE_NAME:
            ((ElementImpl) node).undoableSetAttribute(
                    elementName.attributeName, name, true);
            break;
        }
    }

    /**
     * says whether the name of a node is editable or not
     * 
     * @param node the node considered
     * @return true if the name of this node is editable
     */
    public boolean isNameEditable(TreeElement node) {
        // gets the ElementName for this node
        ElementName elementName = this.getElementName(node);
        // depending on the nameType, returns
        switch (elementName.nameType) {
        case ElementName.TYPE_NAME:
            // it is not editable if the name is the type of the node
            return false;
        case ElementName.VALUE_NAME:
            // returns true if the value exists for an element,
            // and in any case for a text node or comment
            switch (node.getNodeType()) {
            case Node.TEXT_NODE:
            case Node.COMMENT_NODE:
                return true;
            case Node.ELEMENT_NODE:
                return ((ElementImpl) node).getPCDATA() != null;
            default:
                return false;
            }
        case ElementName.ATTRIBUTE_NAME:
            // returns true if the attribute giving its value as name exists
            return ((ElementImpl) node)
                    .getAttributeByName(elementName.attributeName) != null;
        }
        // we should never reach this point
        throw new RuntimeException("Bad value for nameType in ElementName");
    }

    /**
     * returns an ElementName corresponding to the default case for a given
     * node. This is called when the node is not in the dtd or when there is no
     * dtd. This is also called for text nodes, since they are not referenced in
     * the subelement lists
     * 
     * @param node the node considered
     * @return the corresponding default ElementName
     */
    private ElementName getDefaultElementName(Node node) {
        // deals with the case of a text node
        if (node.getNodeType() == Node.TEXT_NODE) { return new ElementName(
                ElementName.VALUE_NAME, null); }
        // else puts the type of the node
        return new ElementName(ElementName.TYPE_NAME, null);
    }

    /**
     * gets the name of a given Node
     * 
     * @param node the node concerned
     * @return its name
     */
    public String getNodeName(TreeElement node) {
        // do we deal with a text node ?
        boolean isTextNode = node.getNodeType() == Node.TEXT_NODE;
        boolean isComment = node.getNodeType() == Node.COMMENT_NODE;
        // gets the ElementName for this node
        ElementName elementName;
        if (isTextNode) {
            elementName = this.getTextNodeName();
        } else if (isComment) {
            elementName = this.getCommentName();
        } else {
            elementName = this.getElementName(node);
        }
        // depending on its value, returns the right value
        switch (elementName.nameType) {
        case ElementName.TYPE_NAME:
            return node.getNodeName();
        case ElementName.VALUE_NAME:
            if (isTextNode || isComment) {
                return node.getNodeValue();
            } else {
                return node.getValue();
            }
        case ElementName.ATTRIBUTE_NAME:
            String result = ((ElementImpl) node)
                    .getAttributeValue(elementName.attributeName);
            if (result != null) {
                return result;
            } else {
                // we're here because the current node does not have the
                // attribute
                // that is supposed to give him its node. In such a case, use
                // the type
                return node.getNodeName();
            }
        }
        // we should never reach this point
        return null;
    }

    /**
     * this describes an element, by listing its attributes and subelements
     */
    private class Element {

        /**
         * says whether this element has a text value or not. a text value is a
         * text subelement, it is not referenced in the subElements hashtable
         */
        private boolean hasValue;
        /**
         * the list of attributes
         */
        private Hashtable attributes;
        /**
         * the list of subelements
         */
        private Hashtable subElements;
        /**
         * this string says where to find a name for this kind of element
         */
        private ElementName elementName;
        /**
         * says whether this element can have any subelement
         */
        private boolean acceptAnySubElement;

        /**
         * default constructor
         */
        public Element() {
            hasValue = false;
            attributes = new Hashtable();
            subElements = new Hashtable();
            elementName = new ElementName(ElementName.TYPE_NAME, "");
            acceptAnySubElement = false;
        }

        /**
         * this tries to put a non stupid default element name
         */
        private void initializeName() {
            if (attributes.get("name") != null) {
                elementName = new ElementName(ElementName.ATTRIBUTE_NAME,
                        "name");
            }
            if (attributes.get("href") != null) {
                elementName = new ElementName(ElementName.ATTRIBUTE_NAME,
                        "href");
            }
        }

        /**
         * adds an attribute to Element
         * 
         * @param name the attribute name
         * @param attribute attribute
         */
        public void addAttribute(String name, Attribute attribute) {
            attributes.put(name, attribute);
        }

        /**
         * adds a subelement to Element
         * 
         * @param name the subelement name
         */
        public void addSubElement(String name) {
            if (!acceptAnySubElement) {
                subElements.put(name, new SubElement());
            }
        }

        /**
         * adds SubElements to Element from an item
         * 
         * @param item the item
         */
        public void addItem(DTDItem item) {
            if (item instanceof DTDAny) {
                anySubElement();
            } else if (item instanceof DTDEmpty) {
                // does nothing
            } else if (item instanceof DTDName) {
                addSubElement(((DTDName) item).getValue());
            } else if (item instanceof DTDPCData) {
                hasValue = true;
            } else if (item instanceof DTDChoice || item instanceof DTDMixed
                    || item instanceof DTDSequence) {
                addContainerItem((DTDContainer) item);
            }
        }

        /**
         * Called if this Element can have any subelement existing
         */
        public void anySubElement() {
            subElements.clear();
            acceptAnySubElement = true;
        }

        /**
         * initializes the subelements table in case of an element accepting any
         * subelements
         */
        private void initializeAnySubElements() {
            Enumeration enm = XmlDisplay.this.elements.keys();
            while (enm.hasMoreElements()) {
                subElements.put(enm.nextElement(), new SubElement());
            }
        }

        /**
         * adds SubElements to Element from a DTDContainer item
         * 
         * @param item the DTDContainer item
         */
        private void addContainerItem(DTDContainer item) {
            DTDItem[] items = item.getItems();
            for (int i = 0; i < items.length; i++) {
                addItem(items[i]);
            }
        }

        /**
         * says whether this component could be seen as an attribute or not To
         * be seen as an attribute means that it only has one subelement (and no
         * attribute) and that this subelement is a text one, containing a value
         * 
         * @return true it it could be seen as an attribute
         */
        public boolean isAttributeLike() {
            // it must have no attributes
            if (attributes.size() == 0) {
                // it must have no subelement
                // it meens that the only contain of this element is string
                // values that could be displayed as attributes
                // so returns the hasValue value
                if (subElements.size() == 0) { return hasValue; }
            }
            // else, it is not attributeLike
            return false;
        }

        /**
         * says whether this element has a text value or not.
         * 
         * @return true if this element has a text value
         */
        public boolean hasValue() {
            return hasValue;
        }

        /**
         * sets the elementName of this Element
         * 
         * @param elementName the new elementName
         */
        public void setElementName(ElementName elementName) {
            this.elementName = elementName;
        }

        /**
         * gets the elementName of this Element
         * 
         * @return the elementName
         */
        public ElementName getElementName() {
            return this.elementName;
        }
    }

    /**
     * this describes a displayed item
     */
    private abstract class DisplayedItem {

        /**
         * true if it should be displayed
         */
        private boolean displayed;

        /**
         * constructor
         */
        public DisplayedItem() {
            displayed = true;
        }

        /**
         * says whether this item should be displayed or not return true if it
         * should be displayed
         */
        public boolean getDisplayed() {
            return displayed;
        }

        /**
         * allows to display or to hide this item
         * 
         * @param displayed if true, the item will be displayed in the future,
         *        if false, it will be hidden
         */
        public void setDisplayed(boolean displayed) {
            this.displayed = displayed;
        }
    }

    /**
     * This describes an attribute
     */
    private class Attribute extends DisplayedItem {

        /**
         * the modifier of the attribute (required, implied, fixed, value)
         */
        private DTDDecl modifier;
        /**
         * the possible values for this attribute null if this attribute can
         * take whatever value
         */
        private String defaultValue;
        /**
         * the possible values for this attribute null if this attribute can
         * take whatever value
         */
        private String[] possibleValues;

        /**
         * constructor
         * 
         * @param values the possible values for this attribute
         * @param modifier the modifier of the attribute (required, implied,
         *        fixed, value)
         * @param defaultValue the default value for this attribute, or null if
         *        none was given
         */
        public Attribute(String[] values, DTDDecl modifier, String defaultValue) {
            super();
            possibleValues = values;
            this.modifier = modifier;
            if (defaultValue != null) {
                // if there is a default value, use it
                this.defaultValue = defaultValue;
            } else {
                // if there is none, try to use a valid value if all values
                // are not allowed
                if (values == null) {
                    this.defaultValue = "aValue";
                } else {
                    this.defaultValue = values[0];
                }
            }
        }

        /**
         * gets the possible values of this attribute
         * 
         * @return the possible values for this attribute or null if any value is
         *        possible
         */
        public String[] getPossibleValues() {
            return possibleValues;
        }

        /**
         * says whether this attribute is required or not
         */
        public boolean isRequired() {
            return modifier == DTDDecl.REQUIRED;
        }

        /**
         * says whether this attribute is editable or not
         */
        public boolean isEditable() {
            return modifier != DTDDecl.FIXED;
        }

        /**
         * gives the default value for this attribute
         * 
         * @return the default value for this attribute
         */
        public String getDefaultValue() {
            return defaultValue;
        }
    }

    /**
     * this describes a subElement
     */
    private class SubElement extends DisplayedItem {

        /**
         * true if it should be displayed as an attribute of its parent
         */
        private boolean displayedAsAttr;

        /**
         * constructor
         */
        public SubElement() {
            super();
            displayedAsAttr = false;
        }

        /**
         * says whether this attribute should be displayed as an attribute of
         * its parent or not return true if it should be displayed as an
         * attribute of its parent
         */
        public boolean getDisplayedAsAttr() {
            return displayedAsAttr;
        }

        /**
         * allows to display or not this element as an attribute for its parent
         * 
         * @param displayed if true, the attribute will be displayed as an
         *        attribute of its parent
         */
        public void setDisplayedAsAttr(boolean displayed) {
            this.displayedAsAttr = displayed;
        }
    }

    /**
     * gets the internal dtd of this XmlDisplay
     * 
     * @return the internal DTD of this XmlDisplay
     */
    public DTD getInternalDtd() {
        return internalDtd;
    }

    /**
     * gets the DtdFile associated with this XmlDisplay. This contains the
     * external dtd of this XmlDisplay.
     * 
     * @return the DtdFile associated with this XmlDisplay
     */
    public DtdFile getDtdFile() {
        return dtdFile;
    }

}
